@@ -73,6 +73,15 @@ static GObjectClass *parent_class = NULL;
 static void cd_recorder_class_init (CDRecorderClass *class);
 static void cd_recorder_init       (CDRecorder      *cdrecorder);
 
+static int cd_recorder_write_iso_cdrecord (CDRecorder *cdrecorder,
+		CDDrive *recorder, const char *filename,
+		int speed, gboolean eject, gboolean blank,
+		gboolean music, gboolean dummy_write);
+static int cd_recorder_write_iso_growisofs (CDRecorder *cdrecorder,
+		CDDrive *recorder, const char *filename,
+		int speed, gboolean eject, gboolean blank,
+		gboolean music, gboolean dummy_write);
+
 GType
 cd_recorder_get_type (void)
 {
@@ -100,14 +109,17 @@ cd_recorder_get_type (void)
 }
 
 static gboolean
-cd_needs_dvd_writer (CDDrive *recorder, const char *filename)
+cd_needs_growisofs (CDDrive *recorder)
 {
-	if (!(recorder->type & CDDRIVE_TYPE_DVD_RAM_RECORDER)
-			|| !(recorder->type & CDDRIVE_TYPE_DVD_RW_RECORDER)) {
-		return FALSE;
-	}
+	CDMediaType type;
 
-	/* XXX check the type of media to see if TRUE */
+	type = guess_media_type (recorder->device);
+	if (type == CD_MEDIA_TYPE_DVDR
+			|| type == CD_MEDIA_TYPE_DVDRW
+			|| type == CD_MEDIA_TYPE_DVD_PLUS_R
+			|| type == CD_MEDIA_TYPE_DVD_PLUS_RW) {
+		return TRUE;
+	}
 
 	return FALSE;
 }
@@ -138,7 +150,72 @@ cd_recorder_insert_cd_retry (CDRecorder *cdrecorder, gboolean is_reload)
 	}
 }
 
-static gboolean  
+static gboolean
+growisofs_stdout_read (GIOChannel   *source,
+		      GIOCondition  condition,
+		      gpointer      data)
+{
+	CDRecorder *cdrecorder = (CDRecorder *) data;
+	char *line;
+	char buf[1];
+	GIOStatus status;
+
+	status = g_io_channel_read_line (source,
+					 &line, NULL, NULL, NULL);
+
+#ifdef DEBUG_CDRECORD
+	if (line) {
+		g_print ("growisofs stdout: %s", line);
+	}
+#endif
+	if (status == G_IO_STATUS_NORMAL) {
+		double percent;
+		int perc_1, perc_2;
+		long long tmp;
+
+		if (cdrecorder->priv->line) {
+			g_string_append (cdrecorder->priv->line, line);
+			g_free (line);
+			line = g_string_free (cdrecorder->priv->line, FALSE);
+			cdrecorder->priv->line = NULL;
+		}
+
+		if (sscanf (line, "%10lld/%lld ( %2d.%1d%%)", &tmp, &tmp, &perc_1, &perc_2) == 4) {
+			if (!cdrecorder->priv->changed_text) {
+				g_signal_emit (G_OBJECT (cdrecorder),
+					       cd_recorder_table_signals[ACTION_CHANGED], 0,
+					       WRITING_CD);
+			}
+
+			percent = (perc_1 + ((float) perc_2 / 10)) / 100;
+
+			g_signal_emit (G_OBJECT (cdrecorder),
+				       cd_recorder_table_signals[PROGRESS_CHANGED], 0,
+				       percent);
+		} else if (strstr (line, "About to execute") != NULL) {
+			g_signal_emit (G_OBJECT (cdrecorder),
+				       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
+				       cancel_cdrecord, TRUE);
+		}
+
+		g_free (line);
+	} else if (status == G_IO_STATUS_AGAIN) {
+		/* A non-terminated line was read, read the data into the buffer. */
+		status = g_io_channel_read_chars (source, buf, 1, NULL, NULL);
+		if (status == G_IO_STATUS_NORMAL) {
+			if (cdrecorder->priv->line == NULL) {
+				cdrecorder->priv->line = g_string_new (NULL);
+			}
+			g_string_append_c (cdrecorder->priv->line, buf[0]);
+		}
+	} else if (status == G_IO_STATUS_EOF) {
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean
 cdrecord_stdout_read (GIOChannel   *source,
 		      GIOCondition  condition,
 		      gpointer      data)
@@ -229,6 +306,62 @@ cdrecord_stdout_read (GIOChannel   *source,
 	return TRUE;
 }
 
+static gboolean
+growisofs_stderr_read (GIOChannel   *source,
+		      GIOCondition  condition,
+		      gpointer      data)
+{
+	CDRecorder *cdrecorder = (CDRecorder *) data;
+	char *line;
+	GIOStatus status;
+
+	status = g_io_channel_read_line (source,
+					 &line, NULL, NULL, NULL);
+#ifdef DEBUG_CDRECORD
+	if (line) {
+		g_print ("growisofs stderr: %s", line);
+	}
+#endif
+
+	/* TODO: Handle errors */
+	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
+		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
+		if (strstr (line, "unsupported MMC profile") != NULL || (strstr (line, "already carries isofs") != NULL && strstr (line, "FATAL:") != NULL)) {
+			g_signal_emit (G_OBJECT (cdrecorder),
+				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
+				       FALSE);
+			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+		} else if (strstr (line, "unable to open") != NULL || strstr (line, "unable to stat") != NULL) {
+			/* This fits the "open64" and "open"-like messages */
+			cdrecorder->priv->last_error = g_strdup (_("The recorder could not be accessed"));
+			cdrecorder->priv->result = RESULT_ERROR;
+		} else if (strstr (line, "not enough space available") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("Not enough space available on the disc"));
+			cdrecorder->priv->result = RESULT_ERROR;
+		} else if (strstr (line, "blocks are free") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
+			cdrecorder->priv->result = RESULT_ERROR;
+		} else if (strstr (line, "flushing cache") != NULL) {
+			g_signal_emit (G_OBJECT (cdrecorder),
+				       cd_recorder_table_signals[ACTION_CHANGED], 0,
+				       FIXATING_CD);
+			cdrecorder->priv->result = RESULT_FINISHED;
+		} else if (strstr (line, ":-(") != NULL || strstr (line, "FATAL") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("Unhandled error, aborting"));
+			cdrecorder->priv->result = RESULT_ERROR;
+		}
+	} else if (status == G_IO_STATUS_EOF) {
+		if (!cdrecorder->priv->expect_cdrecord_to_die) {
+			g_main_loop_quit (cdrecorder->priv->loop);
+		}
+		return FALSE;
+	} else {
+		g_print ("cdrecord stderr read failed, status: %d\n", status);
+	}
+
+	return TRUE;
+}
+
 static gboolean  
 cdrecord_stderr_read (GIOChannel   *source,
 		      GIOCondition  condition,
@@ -268,7 +401,7 @@ cdrecord_stderr_read (GIOChannel   *source,
 				       FALSE);
 			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
 		}
-		
+
 		g_free (line);
 	} else if (status == G_IO_STATUS_EOF) {
 		if (!cdrecorder->priv->expect_cdrecord_to_die) {
@@ -291,6 +424,166 @@ cd_recorder_write_iso (CDRecorder *cdrecorder,
 		       gboolean blank,
 		       gboolean music,
 		       gboolean dummy_write)
+{
+	if (cd_needs_growisofs (recorder)) {
+		return cd_recorder_write_iso_growisofs (cdrecorder,
+				recorder, filename, speed, eject, blank,
+				music, dummy_write);
+	} else {
+		return cd_recorder_write_iso_cdrecord (cdrecorder,
+				recorder, filename, speed, eject, blank,
+				music, dummy_write);
+	}
+}
+
+static int
+cd_recorder_write_iso_growisofs (CDRecorder *cdrecorder,
+				 CDDrive *recorder,
+				 const char *filename,
+				 int speed,
+				 gboolean eject,
+				 gboolean blank,
+				 gboolean music,
+				 gboolean dummy_write)
+{
+	const char *argv[200]; /* Shouldn't need more than 200 arguments */
+	char *speed_str, *dev_str;
+	int stdout_pipe, stderr_pipe;
+	guint stdout_tag, stderr_tag;
+	GIOChannel *channel;
+	GError *error;
+	int i;
+
+	i = 0;
+
+	argv[i++] = "growisofs";
+	speed_str = g_strdup_printf ("-speed=%d", speed);
+	if (speed != 0) {
+		argv[i++] = speed_str;
+	}
+
+	argv[i++] = "-dvd-compat";
+
+	/* Weird, innit? We tell growisofs we have a tty so it ignores
+	 * the fact that the DVD+ has an ISO fs already */
+	if (blank) {
+		argv[i++] = "-use-the-force-luke=tty";
+	}
+
+	argv[i++] = "-Z";
+
+	dev_str = g_strdup_printf ("%s=%s", recorder->device, filename);
+	argv[i++] = dev_str;
+
+	argv[i++] = NULL;
+
+	cdrecorder->priv->cdr_stderr = NULL;
+retry_growisofs:
+	cdrecorder->priv->result = RESULT_ERROR;
+	cdrecorder->priv->expect_cdrecord_to_die = FALSE;
+	cdrecorder->priv->line = NULL;
+	if (cdrecorder->priv->cdr_stderr != NULL) {
+		g_string_truncate (cdrecorder->priv->cdr_stderr, 0);
+	} else {
+		cdrecorder->priv->cdr_stderr = g_string_new (NULL);
+	}
+
+	g_signal_emit (G_OBJECT (cdrecorder),
+			cd_recorder_table_signals[ACTION_CHANGED], 0,
+			PREPARING_TO_WRITE_CD);
+	cdrecorder->priv->changed_text = FALSE;
+	g_signal_emit (G_OBJECT (cdrecorder),
+			cd_recorder_table_signals[PROGRESS_CHANGED], 0,
+			0.0);
+	g_signal_emit (G_OBJECT (cdrecorder),
+			cd_recorder_table_signals[ANIMATION_CHANGED], 0,
+			TRUE);
+	g_signal_emit (G_OBJECT (cdrecorder),
+			cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
+			cancel_cdrecord, FALSE);
+
+#ifdef DEBUG_CDRECORD
+	g_print ("launching command: ");
+	for (i = 0; argv[i] != NULL; i++) {
+		g_print ("%s ", argv[i]);
+	}
+	g_print ("\n");
+#endif
+
+	error = NULL;
+	if (!g_spawn_async_with_pipes  (NULL,
+				        (char **)argv,
+					NULL,
+					G_SPAWN_SEARCH_PATH,
+					NULL, NULL,
+					&cdrecorder->priv->pid,
+					&cdrecorder->priv->cdr_stdin,
+					&stdout_pipe,
+					&stderr_pipe,
+					&error)) {
+		g_warning ("growisofs command failed: %s\n", error->message);
+		for (i = 0; argv[i] != NULL; i++) {
+			g_print ("%s ", argv[i]);
+		}
+		g_print ("\n");
+		g_error_free (error);
+		/* TODO: Better error handling */
+	} else {
+		/* Make sure we don't block on a read. */
+		fcntl (stdout_pipe, F_SETFL, O_NONBLOCK);
+		fcntl (stdout_pipe, F_SETFL, O_NONBLOCK);
+
+		cdrecorder->priv->loop = g_main_loop_new (NULL, FALSE);
+
+		channel = g_io_channel_unix_new (stdout_pipe);
+		g_io_channel_set_encoding (channel, NULL, NULL);
+		stdout_tag = g_io_add_watch (channel,
+					     (G_IO_IN | G_IO_HUP | G_IO_ERR),
+					     growisofs_stdout_read,
+					     cdrecorder);
+		g_io_channel_unref (channel);
+		channel = g_io_channel_unix_new (stderr_pipe);
+		g_io_channel_set_encoding (channel, NULL, NULL);
+		stderr_tag = g_io_add_watch (channel,
+					     (G_IO_IN | G_IO_HUP | G_IO_ERR),
+					     growisofs_stderr_read,
+					     cdrecorder);
+		g_io_channel_unref (channel);
+
+		g_signal_emit (G_OBJECT (cdrecorder),
+			       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
+			       cancel_cdrecord, FALSE);
+
+		g_main_loop_run (cdrecorder->priv->loop);
+		g_main_loop_unref (cdrecorder->priv->loop);
+
+		g_source_remove (stdout_tag);
+		g_source_remove (stderr_tag);
+
+		if (cdrecorder->priv->result == RESULT_RETRY) {
+			goto retry_growisofs;
+		}
+	}
+
+	g_free (speed_str);
+	g_free (dev_str);
+
+	g_signal_emit (G_OBJECT (cdrecorder),
+			cd_recorder_table_signals[ANIMATION_CHANGED], 0,
+			FALSE);
+
+	return cdrecorder->priv->result;
+}
+
+static int
+cd_recorder_write_iso_cdrecord (CDRecorder *cdrecorder,
+		       CDDrive *recorder,
+		       const char *filename,
+		       int speed,
+		       gboolean eject,
+		       gboolean blank,
+		       gboolean music,
+		       gboolean dummy_write)
 {
 	const char *argv[200]; /* Shouldn't need more than 200 arguments */
 	char *speed_str, *dev_str;
@@ -299,17 +592,11 @@ cd_recorder_write_iso (CDRecorder *cdrecorder,
 	GIOChannel *channel;
 	GError *error;
 	int i, saved_i;
-	gboolean needs_dvd_writer;
 
 	i = 0;
 	saved_i = 0;
 
-	needs_dvd_writer = cd_needs_dvd_writer (recorder, filename);
-	if (needs_dvd_writer) {
-		argv[i++] = "dvdrecord";
-	} else {
-		argv[i++] = "cdrecord";
-	}
+	argv[i++] = "cdrecord";
 
 	speed_str = g_strdup_printf ("speed=%d", speed);
 	if (speed != 0) {
@@ -338,10 +625,6 @@ cd_recorder_write_iso (CDRecorder *cdrecorder,
 		argv[i++] = filename;
 	}
 
-	if (needs_dvd_writer) {
-		argv[i++] = "-dao";
-	}
-
 	argv[i++] = NULL;
 
 	cdrecorder->priv->cdr_stderr = NULL;
@@ -369,6 +652,14 @@ cd_recorder_write_iso (CDRecorder *cdrecorder,
 		       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
 		       cancel_cdrecord, FALSE);
 
+#ifdef DEBUG_CDRECORD
+	g_print ("launching command: ");
+	for (i = 0; argv[i] != NULL; i++) {
+		g_print ("%s ", argv[i]);
+	}
+	g_print ("\n");
+#endif
+
 	error = NULL;
 	if (!g_spawn_async_with_pipes  (NULL,
 					(char **)argv,
@@ -389,7 +680,7 @@ cd_recorder_write_iso (CDRecorder *cdrecorder,
 		fcntl (stderr_pipe, F_SETFL, O_NONBLOCK);
 
 		cdrecorder->priv->loop = g_main_loop_new (NULL, FALSE);
-	
+
 		channel = g_io_channel_unix_new (stdout_pipe);
 		g_io_channel_set_encoding (channel, NULL, NULL);
 		stdout_tag = g_io_add_watch (channel, 
